/*
 * Disassembler for warcraft bytecode
 * Input:
 * 	program listing
 * 	string table
 * 	flow list
 *  range of bytecode to disassemble
 */

#include <stdio.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <vector>

using namespace std;

#include "libbs.h"
#include "opcode.h"

FILE *bytecodelog=NULL;
string op_unknown = "unknown";

string optypenames[] = 
{
	op_unknown,
	op_unknown,
	"jump",
	"func_start",
	"func_end",
	"local",
	"global",
	"constant",
	"pop",
	op_unknown,
	op_unknown, // 10
	"clean stack",
	"literal",
	"retval",
	"getvar",
	"code",
	"getvar[]",
	"setvar",
	"setvar[]",
	"push",
	"setreg",
	"callnative",
	"callfunc",
	"typecast",
	op_unknown,
	op_unknown,
	"==",
	op_unknown,
	"<=",
	"<",
	">=",
	">",
	"add",
	"subtract",
	"multiply",
	"divide",
	op_unknown,
	"negate",
	"bnot",
	"return",
	"jump dest",
	"jump if true",
	"jump if false",
	op_unknown,
	op_unknown,
	op_unknown
};

string TYPENAME[] = {
	"unknown","unknown","null","code","integer","real","string","handle","boolean"
};


vector<opcode> oplist;
/*
 * C style binary IO
 */
void buildoplist(const char *filename)
{
	try {
		FILE *opfile = fopen(filename,"r");
		if(!opfile) throw("Could not open bytecode listing");
		int n;
		int read = fread(&n,sizeof(int),1,opfile);
		if(read != 1) throw("Couldn't read from bytecode listing");
		oplist.resize(n);
		opcode op;
		for(int i=0;i<n;i++) {
			fread(&op,sizeof(opcode),1,opfile);
			oplist[i] = op;
		}
	} catch(string s) {
		cerr << "Error: " << s << endl;
		cerr << "Dying" << endl;
		exit(1);
	}
}


bool havesymtable=false;
vector<string> symtable;
void buildsymtable(const char *fn)
{
	try {
		FILE *symf = fopen(fn,"r");
		unsigned i;
		char buf[16384];
		while(fscanf(symf,"%d%s",&i,buf) == 2) {
			if(symtable.size() <= i) symtable.resize(i+1);
			symtable[i] = buf;
		}
		havesymtable = true;
	} catch(string s) {
		cerr << "Error: " << s << endl;
		cerr << "Proceeding with no sym table" << endl;
		havesymtable = false;
	}
}

string getsym(const int x)
{
	if(havesymtable)
		return symtable[x];
	else
		return "nosym";
}

bool havefntable = false;
vector<int> fntable;
void buildfntable()
{
	try {
		fntable.resize(oplist.size());
		if(oplist[0].optype != OP_FUNCTION)	throw("Op listing does not begin with a function!  Corrupt?");
		int fnnamesym = oplist[0].arg;

		for(unsigned i=0;i<oplist.size();i++) {
			if(oplist[i].optype == OP_FUNCTION) {
				fnnamesym = oplist[i].arg;	//look up with getsym()
			}
			fntable[i] = fnnamesym;
		}
		havefntable = true;
	} catch(const char *s) {
		cerr << "Error: " << s << endl;
		cerr << "Proceeding with no function table" << endl;
		havefntable = false;
	}
	
}
string getfn(const int opcodeoffset)
{
	if(havefntable)
		return getsym(fntable[opcodeoffset]);
	else
		return "";
}

string logvtype(int x)		//needs bounds checking
{
	if(x > 8) {
		return ":"+TYPENAME[x-8]+" array";
	} else {
		return ":"+TYPENAME[x];
	}
}

void logbc(ofstream &out, int opindex)
{
	string line = "";
	opcode op = oplist[opindex];
	line += getfn(opindex) + " :: ";
	line += op.r1;
	switch(op.optype) {
		case OP_GETARRAY:
			line += logvtype(op.r1);
	}
	line += " " + op.r2;
	switch(op.optype) {
		case OP_LITERAL:
		case OP_GETVAR:
		case OP_CODE:
			line += logvtype(op.r2);
	}
	line += " " + op.r3;
	switch(op.optype) {
		case OP_LOCAL:
		case OP_GLOBAL:
		case OP_CONSTANT:
		case OP_POPFUNCARG:
			line += logvtype(op.r3);
	}
	line += " ";
	line += "0x";
	line += op.optype+":"+optypenames[op.optype];
	line += " "+op.arg;
	switch(op.optype) {
		case OP_LITERAL:		//if it's J_STRING, follow through
			if (op.r2 != J_STRING) break;
		case OP_NATIVE:
		case OP_JASSCALL:
		case OP_POPFUNCARG:
		case OP_LOCAL:
		case OP_GLOBAL:
		case OP_CONSTANT:
		case OP_SETVAR:
		case OP_GETVAR:
		case OP_SETARRAY:
		case OP_GETARRAY:
			line += ":"+getsym(op.arg);
	}
	out << line << "\n";
}

int main(int argc, char **argv)
{
	popen("foo","w");
}
